#brailleDisplayDrivers/braillesense.py
#A part of NonVisual Desktop Access (NVDA)
#This file is covered by the GNU General Public License.
#See the file COPYING for more details.
#Copyright (C) 2010 Gianluca Casalino <gianluca@spazioausili.net>

from ctypes import *
from ctypes.wintypes import *
import braille
import queueHandler
from logHandler import log
from keyboardHandler import KeyboardInputGesture
pressedKeys=set()
releasedKeys=set()
BS_KEYPRESSED = 0x01
BS_KEYRELEASED = 0x02
BS_CURSORROUTING = 0x00
LRESULT=c_long
HCURSOR=c_long

#MAP OF KEYS
DOT1_KEY = 1
DOT2_KEY = 2
DOT3_KEY = 4
DOT4_KEY = 8
DOT5_KEY = 16
DOT6_KEY = 32
DOT7_KEY = 64
DOT8_KEY = 128
SCROLL_LEFT = 8192
SCROLL_RIGHT = 16384
SPACE = 256
ADVANCE1 = 512
ADVANCE2 = 1024
ADVANCE3 = 2048
ADVANCE4 = 4096

try:
	bsLib = cdll.LoadLibrary("brailleDisplayDrivers\\HanSoneConnect.dll")
except:
	bsLib = None

#Standard window class stuff

WNDPROC=WINFUNCTYPE(LRESULT,HWND,c_uint,WPARAM,LPARAM)

class WNDCLASSEXW(Structure):
	_fields_=[
		('cbSize',c_uint),
		('style',c_uint),
		('lpfnWndProc',WNDPROC),
		('cbClsExtra',c_int),
		('cbWndExtra',c_int),
		('hInstance',HINSTANCE),
		('hIcon',HICON),
		('HCURSOR',HCURSOR),
		('hbrBackground',HBRUSH),
		('lpszMenuName',LPWSTR),
		('lpszClassName',LPWSTR),
		('hIconSm',HICON),
	]

appInstance=windll.kernel32.GetModuleHandleW(None)

nvdaBsBrlWm=windll.user32.RegisterWindowMessageW(u"nvdaBsBrlWm")

@WNDPROC
def nvdaBsBrlWndProc(hwnd,msg,wParam,lParam):
	global pressedKeys, releasedKeys
	if msg == nvdaBsBrlWm and wParam==BS_KEYPRESSED:
		pressedKeys.add(lParam)
		if lParam == SCROLL_LEFT: 
			queueHandler.queueFunction(queueHandler.eventQueue,braille.handler.scrollBack)
		elif lParam == SCROLL_RIGHT:
			queueHandler.queueFunction(queueHandler.eventQueue,braille.handler.scrollForward)
	if msg == nvdaBsBrlWm and wParam==BS_KEYRELEASED:
		releasedKeys.add(lParam)
		if pressedKeys==releasedKeys: 
			_handleKeys(pressedKeys)
			pressedKeys.clear()
			releasedKeys.clear()
	elif msg == nvdaBsBrlWm and wParam == BS_CURSORROUTING:
		queueHandler.queueFunction(queueHandler.eventQueue,braille.handler.routeTo, lParam)
	return windll.user32.DefWindowProcW(hwnd,msg,wParam,lParam)

def _handleKeys(keys):
	if keys == set([DOT4_KEY,DOT5_KEY,SPACE]):
		queueHandler.queueFunction(queueHandler.eventQueue, (KeyboardInputGesture.fromName('tab')).send)
	if keys == set([DOT1_KEY,DOT2_KEY,SPACE]): 
		queueHandler.queueFunction(queueHandler.eventQueue,(KeyboardInputGesture.fromName('shift+tab')).send)
	if keys == set([128]): 
		queueHandler.queueFunction(queueHandler.eventQueue,(KeyboardInputGesture.fromName('enter')).send) 
	if keys == set([256]): 
		queueHandler.queueFunction(queueHandler.eventQueue,(KeyboardInputGesture.fromName('space')).send) 
	if keys == set([256,DOT1_KEY,DOT5_KEY]): 
		queueHandler.queueFunction(queueHandler.eventQueue,(KeyboardInputGesture.fromName('escape')).send) 
	if keys == set([256,DOT1_KEY]): 
		queueHandler.queueFunction(queueHandler.eventQueue,(KeyboardInputGesture.fromName('upArrow')).send)
	if keys == set([256,DOT3_KEY]): 
		queueHandler.queueFunction(queueHandler.eventQueue,(KeyboardInputGesture.fromName('leftArrow')).send) 
	if keys == set([256,DOT6_KEY]): 
		queueHandler.queueFunction(queueHandler.eventQueue,(KeyboardInputGesture.fromName('rightArrow')).send) 
	if keys == set([256,DOT4_KEY]): 
		queueHandler.queueFunction(queueHandler.eventQueue,(KeyboardInputGesture.fromName('downArrow')).send) 

nvdaBsBrlWndCls=WNDCLASSEXW()
nvdaBsBrlWndCls.cbSize=sizeof(nvdaBsBrlWndCls)
nvdaBsBrlWndCls.lpfnWndProc=nvdaBsBrlWndProc
nvdaBsBrlWndCls.hInstance=appInstance
nvdaBsBrlWndCls.lpszClassName=u"nvdaBsBrlWndCls"

class BrailleDisplayDriver(braille.BrailleDisplayDriverWithCursor):
	"""Braille Sense.
	"""
	name = "braillesense"
	description = "Braille Sense"

	@classmethod
	def check(cls):
		return bool(bsLib)

	def __init__(self):
		super(BrailleDisplayDriver, self).__init__()
		self._messageWindowClassAtom=windll.user32.RegisterClassExW(byref(nvdaBsBrlWndCls))
		self._messageWindow=windll.user32.CreateWindowExW(0,self._messageWindowClassAtom,u"nvdaBsBrlWndCls window",0,0,0,0,0,None,None,appInstance,None)
		connected = bsLib.Open("usb",self._messageWindow,nvdaBsBrlWm)
		if connected==0: 
			for n in xrange(1,256):
				port=("com%d"%n)
				if n > 8: port="\\\\.\\"+port
				connected = bsLib.Open(port,self._messageWindow,nvdaBsBrlWm)
				if connected>=1: return
			raise RuntimeError("No display found")
 
	def terminate(self):
		super(BrailleDisplayDriver, self).terminate()
		bsLib.Close()
		windll.user32.DestroyWindow(self._messageWindow)
		windll.user32.UnregisterClassW(self._messageWindowClassAtom,appInstance)

 	def _get_numCells(self):
		return bsLib.GetBSCellCount()

	def _display(self, cells):
		cells = "".join([chr(x) for x in cells])
		bsLib.SendData(cells)
